#include "lcars.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/syscall.h>

static char *shared_buf = (char *) 0x3000000;
static int32_t *shared_ctrl = (int32_t *)0x20000000 ;

int64_t syscall1(int number, int64_t _1) {
        int64_t ret;
        __asm__ volatile(
            "syscall\n\t"
            : "=a"(ret)
            : "a"(number),
              "D"(_1)
            : "memory", "rcx", "r11");
        return ret;
}
 
int64_t syscall3(int number, int64_t _1, int64_t _2, int64_t _3) {
        int64_t ret;
        __asm__ volatile(
            "syscall\n\t"
            : "=a"(ret)
            : "a"(number),
              "D"(_1),
              "S"(_2),
              "d"(_3)
            : "memory", "rcx", "r11");
        return ret;
}
 
#define syscall4(number, _1, _2, _3, _4)      \
    ({                                        \
        int64_t ret;                          \
        register int64_t r10 asm("r10") = _4; \
        __asm__ volatile(                     \
            "syscall\n\t"                     \
            : "=a"(ret)                       \
            : "a"(number),                    \
              "D"(_1),                        \
              "S"(_2),                        \
              "d"(_3),                        \
              "r"(r10)                        \
            : "memory", "rcx", "r11");        \
        ret;                                  \
    })

char * get_shared_buf (void) {
    return shared_buf;
}

int32_t *get_shared_ctrl (void) {
    return shared_ctrl;
}

void _exit(int status) {
    //syscall(SYS_exit, status);
    syscall1(SYS_exit, status);
}



ssize_t write(int fd, const void *buf, size_t len) {
    //return syscall(SYS_write, fd, buf, len);
    return syscall3(SYS_write, fd, (int64_t) buf, len);
}

ssize_t read(int fd, void *buf, size_t len) {
    //return syscall(SYS_read, fd, buf, len);
    return syscall3(SYS_read, fd, (int64_t) buf, len);
}

size_t strlen(const char *str) {
    int len = 0;
    int i;

    for (i = 0; str[i] != 0; i++)
    {
        len++;
    }
    return (len);
}

char *strcpy(char *d, const char *s) {
    char *saved = d;
    while (*s)
    {
        *d++ = *s++;
    }
    *d = 0;
    return saved;
}

int32_t do_request(int32_t type, int32_t arg1, int32_t arg2, int32_t arg3, int32_t arg4)
{
    request req;
    req.type = type;
    req.arg1 = arg1;
    req.arg2 = arg2;
    req.arg3 = arg3;
    req.arg4 = arg4;
    
    int32_t resp;

    //if (write(STDOUT, &req, 20) != 20)
    if (write( STDOUT, &req, 20) != 20)
        _exit(1);

    resp = 0;
    //if ((unsigned int)read(STDIN, &resp, 4) != 4)
    if ((unsigned int)read( STDIN, &resp, 4) != 4)
        _exit(2);
    return resp;
}

int32_t allocate(int32_t len) {

    int64_t result; // rax
    unsigned int v2;       // edi
    bool v3;               // cc

    result = -1;
    if (len <= 0x1000000)
    {
        result = shared_ctrl[0];
        v2 = (len + 0xF) & 0xFFFFFFF0;
        v3 = v2 + shared_ctrl[0] > 0xFFFFFF;
        if (v2 + shared_ctrl[0] <= 0xFFFFFF)
            v2 += shared_ctrl[0];
        if (v3)
            result = 0LL;
        shared_ctrl[0] = v2 & 0xFFFFFF;
    }
    return result;
}

void *memcpy(void *dest, const void *src, size_t n)
{
    // Typecast src and dest addresses to (char *)
    char *csrc = (char *)src;
    char *cdest = (char *)dest;

    // Copy contents of src[] to dest[]
    for (int i = 0; i < n; i++)
        cdest[i] = csrc[i];
}

int get_file(char *name) {
    int fname_len;
    int32_t shared_idx;
    int32_t result;

    int len;
    int fd;
    char buf[1];
    struct iovec iov;
    struct msghdr msg;
    struct cmsghdr *cmsg;
    char cms[CMSG_SPACE(sizeof(int))];



    fname_len = strlen(name) + 1;
    shared_idx = allocate(fname_len);
    strcpy(shared_buf + shared_idx, name);
    result = do_request(5, shared_idx, fname_len, 0, 0);
    if (!result) {
        iov.iov_base = buf;
        iov.iov_len = sizeof(buf);

        msg.msg_name = 0;
        msg.msg_namelen = 0;
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;
        msg.msg_flags = 0;
        msg.msg_control = cms;
        msg.msg_controllen = sizeof cms;

        //len = recvmsg(STDIN, &msg, 0);
        //len = syscall(SYS_recvmsg, STDIN, &msg, 0);
        len = syscall3(SYS_recvmsg, STDIN, (int64_t) &msg, 0);

        if (len < 0)
        {
            _exit(3);
        }

        cmsg = CMSG_FIRSTHDR(&msg);
        memcpy(&fd, CMSG_DATA(cmsg), sizeof(int));
        return fd;
    }

}